module derived_types

  ! Libraries:
  use contour_531, only: polyline
  use shapelib, only: shpfileobject

  implicit none

  private shpfileobject

  real, parameter:: missing_ssh = 1e4 ! flag for undefined SSH
  real, parameter:: missing_speed = 1e4 ! flag for missing max speed

  type, extends(polyline):: ssh_contour
     ! Points of polyline are (longitude, latitude) in rad. Polyline
     ! should be closed, except if it is null.
     real ssh ! in m
     real:: area = - 1e6 ! > 0 if not initialization value, in m2
  end type ssh_contour

  type extremum
     integer:: coord_proj(2) = huge(0) ! coordinates in projection space
     real coord (2) ! longitude and latitude, in rad
     real ssh ! value of ssh, in m
  end type extremum

  type eddy
     type(extremum) extr
     logical cyclone
     type(ssh_contour) out_cont ! outermost contour
     type(ssh_contour) speed_cont ! contour with maximum average azimuthal speed

     real innermost_level
     ! SSH level of the innermost contour around each extremum, in
     ! m. By construction, innermost_level < extremum for a maximum, >
     ! extremum for a minimum.

     real max_speed
     ! Average of azimuthal speed on out_cont or speed_cont, in m
     ! s-1. Positive means counterclockwise circulation. If speed_cont
     ! is not a null ssh contour then max_speed is the speed on
     ! speed_cont.

     integer:: delta_in = huge(0)
     ! Minimum difference in time subscript where there is a direct
     ! predecessor. Huge means there is no direct predecessor.

     integer:: delta_out = huge(0)
     ! Minimum difference in time subscript where there is a direct
     ! successor. Huge means there is no direct succcessor.
  end type eddy

  type snapshot
     type(eddy), allocatable:: list(:) ! (number_extr) "Eddies" at a
     ! given date. In general, there may be both cyclonic and
     ! anticyclonic eddies in the list. Eddies include "eddies"
     ! without an outermost contour, that is, only the extremum is
     ! defined. The subscript value in list is the "identification
     ! number" of the extremum.

     integer number_extr ! number of extrema

     integer, allocatable:: extr_map(:, :)
     ! (1 - copy:nlon + copy, nlat) if the grid is periodic in
     ! longitude, else (nlon, nlat). copy can be max_radius(1) or
     ! dist_lim, depending on the program. At a point of extremum SSH:
     ! identification number or this extremum. 0 at other points.
  end type snapshot

  type ori_slice_handler
     ! Slice of a shapefile collection, for a given orientation
     ! (cyclonic or anticyclonic eddies).

     TYPE(shpfileobject) extremum ! shapefile extremum

     TYPE(shpfileobject) extr_proj
     ! shapefile for coordinates of extrema in projection plane

     TYPE(shpfileobject) outermost ! shapefile outermost_contour
     TYPE(shpfileobject) max_speed ! shapefile max_speed_contour
     integer unit ! ishape_last.txt
     logical cyclone
     logical with_proj ! with shapefile extr_proj

     ! Field identifiers in the DBF files:
     integer extr_ssh, extr_date, extr_eddy_index, extr_speed
     integer extr_proj_date, extr_proj_eddy_index
     integer out_r_eq_area, out_ssh, out_date, out_eddy_index
     integer max_speed_r_eq_area, max_speed_ssh, max_speed_date, &
          max_speed_eddy_index

     character(len = :), allocatable:: dir ! directory
  end type Ori_Slice_Handler

  type shpc_slice_meta
     ! Some metadata of a slice of SHPC
     integer d0 ! first date

     integer, allocatable:: ishape_last(:) ! (d0:)
     ! shape index (0-based) in the collection of shapefiles of the last
     ! shape at a given date index
  end type shpc_slice_meta

contains

  pure type(ssh_contour) function null_ssh_contour()

    use contour_531, only: null_polyline

    !--------------------------------------------------------------

    null_ssh_contour%polyline = null_polyline()
    null_ssh_contour%ssh = missing_ssh
    null_ssh_contour%area = - 1e6

  end function null_ssh_contour

end module derived_types
